@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2009-2010 Nokia Corporation and/or its subsidiary(-ies).
+ *  Copyright (C) 2009-2010  Nokia Corporation and/or its subsidiary(-ies).
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -23,7 +23,6 @@
 #include <config.h>
 #endif
 
-#define _GNU_SOURCE
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -44,7 +43,6 @@
 #include <gisi/pep.h>
 #include <gisi/pipe.h>
 
-#include "isimodem.h"
 #include "isiutil.h"
 #include "gpds.h"
 #include "debug.h"
@@ -535,14 +533,15 @@ static void isi_gprs_activate_primary(struct ofono_gprs_context *gc,
 			|| strlen(ctx->password) >= GPDS_MAX_PASSWORD_LENGTH)
 		goto error;
 
-	strncpy(cd->apn, ctx->apn, GPDS_MAX_APN_STRING_LENGTH);
-	cd->apn[GPDS_MAX_APN_STRING_LENGTH] = '\0';
+	g_strlcpy(cd->apn, ctx->apn, sizeof(cd->apn));
 
-	strncpy(cd->username, ctx->username, GPDS_MAX_USERNAME_LENGTH);
-	cd->username[GPDS_MAX_USERNAME_LENGTH] = '\0';
-
-	strncpy(cd->password, ctx->password, GPDS_MAX_PASSWORD_LENGTH);
-	cd->username[GPDS_MAX_PASSWORD_LENGTH] = '\0';
+	if (ctx->auth_method == OFONO_GPRS_AUTH_METHOD_NONE) {
+		memset(cd->username, 0, sizeof(cd->username));
+		memset(cd->password, 0, sizeof(cd->password));
+	} else {
+		g_strlcpy(cd->username, ctx->username, sizeof(cd->username));
+		g_strlcpy(cd->password, ctx->password, sizeof(cd->password));
+	}
 
 	cd->pep = g_isi_pep_create(cd->idx, NULL, NULL);
 	if (cd->pep == NULL)
@@ -659,20 +658,11 @@ static void isi_gprs_context_remove(struct ofono_gprs_context *gc)
 	g_free(cd);
 }
 
-static struct ofono_gprs_context_driver driver = {
-	.name			= "isimodem",
+static const struct ofono_gprs_context_driver driver = {
 	.probe			= isi_gprs_context_probe,
 	.remove			= isi_gprs_context_remove,
 	.activate_primary	= isi_gprs_activate_primary,
 	.deactivate_primary	= isi_gprs_deactivate_primary,
 };
 
-void isi_gprs_context_init(void)
-{
-	ofono_gprs_context_driver_register(&driver);
-}
-
-void isi_gprs_context_exit(void)
-{
-	ofono_gprs_context_driver_unregister(&driver);
-}
+OFONO_ATOM_DRIVER_BUILTIN(gprs_context, isimodem, &driver)
