@@ -34,14 +34,27 @@
 
 #include "qmi.h"
 #include "nas.h"
+#include "util.h"
 
-#include "qmimodem.h"
 #include "src/common.h"
 
 struct netreg_data {
 	struct qmi_service *nas;
 	struct ofono_network_operator operator;
 	uint8_t current_rat;
+	int lac;
+	int cellid;
+	bool is_roaming;
+	uint16_t event_indication_id;
+	uint16_t signal_info_indication_id;
+	uint16_t system_info_indication_id;
+	uint16_t serving_system_indication_id;
+};
+
+enum roaming_status {
+	ROAMING_STATUS_OFF,
+	ROAMING_STATUS_ON,
+	ROAMING_STATUS_NO_CHANGE,
 };
 
 static bool extract_ss_info_time(
@@ -78,11 +91,12 @@ static bool extract_ss_info_time(
 
 static bool extract_ss_info(struct qmi_result *result, int *status,
 				int *lac, int *cellid, int *tech,
+				enum roaming_status *roaming,
 				struct ofono_network_operator *operator)
 {
 	const struct qmi_nas_serving_system *ss;
 	const struct qmi_nas_current_plmn *plmn;
-	uint8_t i, roaming;
+	uint8_t i, roaming_status;
 	uint16_t value16, len, opname_len;
 	uint32_t value32;
 
@@ -104,10 +118,13 @@ static bool extract_ss_info(struct qmi_result *result, int *status,
 		*tech = qmi_nas_rat_to_tech(ss->radio_if[i]);
 	}
 
+	*roaming = ROAMING_STATUS_NO_CHANGE;
 	if (qmi_result_get_uint8(result, QMI_NAS_RESULT_ROAMING_STATUS,
-								&roaming)) {
-		if (ss->status == 1 && roaming == 0)
-			*status = NETWORK_REGISTRATION_STATUS_ROAMING;
+							&roaming_status)) {
+		if (roaming_status == 0)
+			*roaming = ROAMING_STATUS_ON;
+		else if (roaming_status == 1)
+			*roaming = ROAMING_STATUS_OFF;
 	}
 
 	if (!operator)
@@ -115,10 +132,18 @@ static bool extract_ss_info(struct qmi_result *result, int *status,
 
 	plmn = qmi_result_get(result, QMI_NAS_RESULT_CURRENT_PLMN, &len);
 	if (plmn) {
-		snprintf(operator->mcc, OFONO_MAX_MCC_LENGTH + 1, "%03d",
-						GUINT16_FROM_LE(plmn->mcc));
-		snprintf(operator->mnc, OFONO_MAX_MNC_LENGTH + 1, "%02d",
-						GUINT16_FROM_LE(plmn->mnc));
+		uint16_t mcc = L_LE16_TO_CPU(plmn->mcc);
+		uint16_t mnc = L_LE16_TO_CPU(plmn->mnc);
+
+		if (mcc > 999)
+			mcc = 999;
+
+		if (mnc > 999)
+			mnc = 999;
+
+		snprintf(operator->mcc, OFONO_MAX_MCC_LENGTH + 1, "%03d", mcc);
+		snprintf(operator->mnc, OFONO_MAX_MNC_LENGTH + 1, "%03d", mnc);
+
 		opname_len = plmn->desc_len;
 		if (opname_len > OFONO_MAX_OPERATOR_NAME_LENGTH)
 			opname_len = OFONO_MAX_OPERATOR_NAME_LENGTH;
@@ -128,7 +153,7 @@ static bool extract_ss_info(struct qmi_result *result, int *status,
 		 * plmn-desc. When that happens, libdbus will abort ofono.
 		 * If non-utf-8 characters are detected, use mccmnc string.
 		 */
-		if (g_utf8_validate(plmn->desc, opname_len, NULL)) {
+		if (l_utf8_validate(plmn->desc, opname_len, NULL)) {
 			strncpy(operator->name, plmn->desc, opname_len);
 			operator->name[opname_len] = '\0';
 		} else
@@ -149,28 +174,58 @@ static bool extract_ss_info(struct qmi_result *result, int *status,
 	else
 		*cellid = -1;
 
-	DBG("lac %d cellid %d tech %d", *lac, *cellid, *tech);
+	DBG("roaming %u lac %d cellid %d tech %d", *roaming, *lac, *cellid,
+									*tech);
 
 	return true;
 }
 
+static int remember_ss_info(struct netreg_data *data, int status, int lac,
+					int cellid, enum roaming_status roaming)
+{
+	if (roaming == ROAMING_STATUS_ON)
+		data->is_roaming = true;
+	else if (roaming == ROAMING_STATUS_OFF)
+		data->is_roaming = false;
+
+	if (status == QMI_NAS_REGISTRATION_STATE_REGISTERED) {
+		if (lac >= 0)
+			data->lac = lac;
+		if (cellid >= 0)
+			data->cellid = cellid;
+	} else {
+		data->lac = -1;
+		data->cellid = -1;
+	}
+
+	if (status == QMI_NAS_REGISTRATION_STATE_REGISTERED &&
+							data->is_roaming)
+		status = NETWORK_REGISTRATION_STATUS_ROAMING;
+
+	return status;
+}
+
 static void ss_info_notify(struct qmi_result *result, void *user_data)
 {
 	struct ofono_netreg *netreg = user_data;
 	struct ofono_network_time net_time;
 	struct netreg_data *data = ofono_netreg_get_data(netreg);
 	int status, lac, cellid, tech;
+	enum roaming_status roaming;
 
 	DBG("");
 
 	if (extract_ss_info_time(result, &net_time))
 		ofono_netreg_time_notify(netreg, &net_time);
 
-	if (!extract_ss_info(result, &status, &lac, &cellid, &tech,
+	if (!extract_ss_info(result, &status, &lac, &cellid, &tech, &roaming,
 							&data->operator))
 		return;
 
-	ofono_netreg_status_notify(netreg, status, lac, cellid, tech);
+	status = remember_ss_info(data, status, lac, cellid, roaming);
+
+	ofono_netreg_status_notify(netreg, status, data->lac, data->cellid,
+									tech);
 }
 
 static void get_ss_info_cb(struct qmi_result *result, void *user_data)
@@ -179,6 +234,7 @@ static void get_ss_info_cb(struct qmi_result *result, void *user_data)
 	ofono_netreg_status_cb_t cb = cbd->cb;
 	struct netreg_data *data = cbd->user;
 	int status, lac, cellid, tech;
+	enum roaming_status roaming;
 
 	DBG("");
 
@@ -187,13 +243,16 @@ static void get_ss_info_cb(struct qmi_result *result, void *user_data)
 		return;
 	}
 
-	if (!extract_ss_info(result, &status, &lac, &cellid, &tech,
+	if (!extract_ss_info(result, &status, &lac, &cellid, &tech, &roaming,
 							&data->operator)) {
 		CALLBACK_WITH_FAILURE(cb, -1, -1, -1, -1, cbd->data);
 		return;
 	}
 
-	CALLBACK_WITH_SUCCESS(cb, status, lac, cellid, tech, cbd->data);
+	status = remember_ss_info(data, status, lac, cellid, roaming);
+
+	CALLBACK_WITH_SUCCESS(cb, status, data->lac, data->cellid, tech,
+								cbd->data);
 }
 
 static void qmi_registration_status(struct ofono_netreg *netreg,
@@ -206,13 +265,13 @@ static void qmi_registration_status(struct ofono_netreg *netreg,
 
 	cbd->user = data;
 
-	if (qmi_service_send(data->nas, QMI_NAS_GET_SS_INFO, NULL,
-					get_ss_info_cb, cbd, g_free) > 0)
+	if (qmi_service_send(data->nas, QMI_NAS_GET_SERVING_SYSTEM, NULL,
+					get_ss_info_cb, cbd, l_free) > 0)
 		return;
 
 	CALLBACK_WITH_FAILURE(cb, -1, -1, -1, -1, cbd->data);
 
-	g_free(cbd);
+	l_free(cbd);
 }
 
 static void qmi_current_operator(struct ofono_netreg *netreg,
@@ -250,25 +309,25 @@ static void scan_nets_cb(struct qmi_result *result, void *user_data)
 
 	netlist = ptr;
 
-	num = GUINT16_FROM_LE(netlist->count);
-
+	num = L_LE16_TO_CPU(netlist->count);
 	DBG("found %d operators", num);
-
-	list = g_try_new0(struct ofono_network_operator, num);
-	if (!list) {
-		CALLBACK_WITH_FAILURE(cb, 0, NULL, cbd->data);
-		return;
-	}
-
 	offset = 2;
 
+	list = l_new(struct ofono_network_operator, num);
+
 	for (i = 0; i < num; i++) {
 		const struct qmi_nas_network_info *netinfo = ptr + offset;
+		uint16_t mcc = L_LE16_TO_CPU(netinfo->mcc);
+		uint16_t mnc = L_LE16_TO_CPU(netinfo->mnc);
 
-		snprintf(list[i].mcc, OFONO_MAX_MCC_LENGTH + 1, "%03d",
-						GUINT16_FROM_LE(netinfo->mcc));
-		snprintf(list[i].mnc, OFONO_MAX_MNC_LENGTH + 1, "%02d",
-						GUINT16_FROM_LE(netinfo->mnc));
+		if (mcc > 999)
+			mcc = 999;
+
+		if (mnc > 999)
+			mnc = 999;
+
+		snprintf(list[i].mcc, OFONO_MAX_MCC_LENGTH + 1, "%03d", mcc);
+		snprintf(list[i].mnc, OFONO_MAX_MNC_LENGTH + 1, "%03d", mnc);
 		strncpy(list[i].name, netinfo->desc, netinfo->desc_len);
 		list[i].name[netinfo->desc_len] = '\0';
 
@@ -294,7 +353,7 @@ static void scan_nets_cb(struct qmi_result *result, void *user_data)
 	if (!netrat)
 		goto done;
 
-	if (GUINT16_FROM_LE(netrat->count) != num)
+	if (L_LE16_TO_CPU(netrat->count) != num)
 		goto done;
 
 	for (i = 0; i < num; i++) {
@@ -307,7 +366,7 @@ static void scan_nets_cb(struct qmi_result *result, void *user_data)
 done:
 	CALLBACK_WITH_SUCCESS(cb, num, list, cbd->data);
 
-	g_free(list);
+	l_free(list);
 }
 
 static void qmi_list_operators(struct ofono_netreg *netreg,
@@ -318,13 +377,13 @@ static void qmi_list_operators(struct ofono_netreg *netreg,
 
 	DBG("");
 
-	if (qmi_service_send(data->nas, QMI_NAS_SCAN_NETS, NULL,
-					scan_nets_cb, cbd, g_free) > 0)
+	if (qmi_service_send(data->nas, QMI_NAS_NETWORK_SCAN, NULL,
+					scan_nets_cb, cbd, l_free) > 0)
 		return;
 
 	CALLBACK_WITH_FAILURE(cb, 0, NULL, cbd->data);
 
-	g_free(cbd);
+	l_free(cbd);
 }
 
 static void register_net_cb(struct qmi_result *result, void *user_data)
@@ -332,6 +391,7 @@ static void register_net_cb(struct qmi_result *result, void *user_data)
 	struct cb_data *cbd = user_data;
 	ofono_netreg_register_cb_t cb = cbd->cb;
 	uint16_t error;
+	int cme_error;
 
 	DBG("");
 
@@ -341,7 +401,8 @@ static void register_net_cb(struct qmi_result *result, void *user_data)
 			goto done;
 		}
 
-		CALLBACK_WITH_FAILURE(cb, cbd->data);
+		cme_error = qmi_error_to_ofono_cme(error);
+		CALLBACK_WITH_CME_ERROR(cb, cme_error, cbd->data);
 		return;
 	}
 
@@ -360,19 +421,14 @@ static void qmi_register_auto(struct ofono_netreg *netreg,
 
 	param = qmi_param_new_uint8(QMI_NAS_PARAM_REGISTER_ACTION,
 					QMI_NAS_REGISTER_ACTION_AUTO);
-	if (!param)
-		goto error;
 
-	if (qmi_service_send(data->nas, QMI_NAS_REGISTER_NET, param,
-					register_net_cb, cbd, g_free) > 0)
+	if (qmi_service_send(data->nas, QMI_NAS_NETWORK_REGISTER, param,
+					register_net_cb, cbd, l_free) > 0)
 		return;
 
 	qmi_param_free(param);
-
-error:
 	CALLBACK_WITH_FAILURE(cb, cbd->data);
-
-	g_free(cbd);
+	l_free(cbd);
 }
 
 static void qmi_register_manual(struct ofono_netreg *netreg,
@@ -388,8 +444,6 @@ static void qmi_register_manual(struct ofono_netreg *netreg,
 
 	param = qmi_param_new_uint8(QMI_NAS_PARAM_REGISTER_ACTION,
 					QMI_NAS_REGISTER_ACTION_MANUAL);
-	if (!param)
-		goto error;
 
 	info.mcc = atoi(mcc);
 	info.mnc = atoi(mnc);
@@ -398,16 +452,13 @@ static void qmi_register_manual(struct ofono_netreg *netreg,
 	qmi_param_append(param, QMI_NAS_PARAM_REGISTER_MANUAL_INFO,
 						sizeof(info), &info);
 
-	if (qmi_service_send(data->nas, QMI_NAS_REGISTER_NET, param,
-					register_net_cb, cbd, g_free) > 0)
+	if (qmi_service_send(data->nas, QMI_NAS_NETWORK_REGISTER, param,
+					register_net_cb, cbd, l_free) > 0)
 		return;
 
 	qmi_param_free(param);
-
-error:
 	CALLBACK_WITH_FAILURE(cb, cbd->data);
-
-	g_free(cbd);
+	l_free(cbd);
 }
 
 static int dbm_to_strength(int8_t dbm)
@@ -464,13 +515,27 @@ static void qmi_signal_strength(struct ofono_netreg *netreg,
 
 	DBG("");
 
-	if (qmi_service_send(data->nas, QMI_NAS_GET_RSSI, NULL,
-					get_rssi_cb, cbd, g_free) > 0)
+	if (qmi_service_send(data->nas, QMI_NAS_GET_SIGNAL_STRENGTH, NULL,
+					get_rssi_cb, cbd, l_free) > 0)
 		return;
 
 	CALLBACK_WITH_FAILURE(cb, -1, cbd->data);
 
-	g_free(cbd);
+	l_free(cbd);
+}
+
+static void system_info_notify(struct qmi_result *result, void *user_data)
+{
+	DBG("");
+
+	qmi_result_print_tlvs(result);
+}
+
+static void signal_info_notify(struct qmi_result *result, void *user_data)
+{
+	DBG("");
+
+	qmi_result_print_tlvs(result);
 }
 
 static void event_notify(struct qmi_result *result, void *user_data)
@@ -508,20 +573,66 @@ static void event_notify(struct qmi_result *result, void *user_data)
 	}
 }
 
-static void set_event_cb(struct qmi_result *result, void *user_data)
+static void register_indications_cb(struct qmi_result *result,
+							void *user_data)
 {
 	struct ofono_netreg *netreg = user_data;
 	struct netreg_data *data = ofono_netreg_get_data(netreg);
 
-	DBG("");
+	if (qmi_result_set_error(result, NULL)) {
+		ofono_netreg_remove(netreg);
+		return;
+	}
 
 	ofono_netreg_register(netreg);
 
-	qmi_service_register(data->nas, QMI_NAS_EVENT,
+	data->event_indication_id =
+		qmi_service_register(data->nas, QMI_NAS_EVENT_REPORT,
 					event_notify, netreg, NULL);
 
-	qmi_service_register(data->nas, QMI_NAS_SS_INFO_IND,
+	data->serving_system_indication_id =
+		qmi_service_register(data->nas,
+					QMI_NAS_SERVING_SYSTEM_INDICATION,
 					ss_info_notify, netreg, NULL);
+
+	data->system_info_indication_id =
+		qmi_service_register(data->nas,
+					QMI_NAS_SYSTEM_INFO_INDICATION,
+					system_info_notify, netreg, NULL);
+
+	data->signal_info_indication_id =
+		qmi_service_register(data->nas,
+					QMI_NAS_SIGNAL_INFO_INDICATION,
+					signal_info_notify, netreg, NULL);
+}
+
+static void set_event_report_cb(struct qmi_result *result, void *user_data)
+{
+	struct ofono_netreg *netreg = user_data;
+	struct netreg_data *data = ofono_netreg_get_data(netreg);
+	static const uint8_t PARAM_SERVING_SYSTEM_EVENTS = 0x13;
+	static const uint8_t PARAM_SYSTEM_INFO = 0x18;
+	static const uint8_t PARAM_SIGNAL_INFO = 0x19;
+	struct qmi_param *param;
+
+	DBG("");
+
+	if (qmi_result_set_error(result, NULL))
+		goto error;
+
+	param = qmi_param_new();
+
+	qmi_param_append_uint8(param, PARAM_SERVING_SYSTEM_EVENTS, 0x01);
+	qmi_param_append_uint8(param, PARAM_SYSTEM_INFO, 0x01);
+	qmi_param_append_uint8(param, PARAM_SIGNAL_INFO, 0x01);
+
+	if (qmi_service_send(data->nas, QMI_NAS_REGISTER_INDICATIONS, param,
+				register_indications_cb, netreg, NULL) > 0)
+		return;
+
+	qmi_param_free(param);
+error:
+	ofono_netreg_remove(netreg);
 }
 
 static void create_nas_cb(struct qmi_service *service, void *user_data)
@@ -529,36 +640,37 @@ static void create_nas_cb(struct qmi_service *service, void *user_data)
 	struct ofono_netreg *netreg = user_data;
 	struct netreg_data *data = ofono_netreg_get_data(netreg);
 	struct qmi_param *param;
-	struct qmi_nas_param_event_signal_strength ss = { .report = 0x01,
-				.count = 5, .dbm[0] = -55, .dbm[1] = -65,
-				.dbm[2] = -75, .dbm[3] = -85, .dbm[4] = -95 };
+	static const uint8_t PARAM_REPORT_SIGNAL_STRENGTH = 0x10;
+	static const uint8_t PARAM_REPORT_RF_INFO = 0x11;
+	struct {
+		uint8_t report;					/* bool */
+		uint8_t count;
+		int8_t dbm[5];
+	} __attribute__((__packed__)) ss = { .report = 0x01,
+			.count = 5, .dbm[0] = -55, .dbm[1] = -65,
+			.dbm[2] = -75, .dbm[3] = -85, .dbm[4] = -95 };
 
 	DBG("");
 
 	if (!service) {
 		ofono_error("Failed to request NAS service");
-		ofono_netreg_remove(netreg);
-		return;
+		goto error;
 	}
 
-	data->nas = qmi_service_ref(service);
+	data->nas = service;
 
 	param = qmi_param_new();
-	if (!param)
-		goto done;
 
-	qmi_param_append(param, QMI_NAS_PARAM_REPORT_SIGNAL_STRENGTH,
-							sizeof(ss), &ss);
-	qmi_param_append_uint8(param, QMI_NAS_PARAM_REPORT_RF_INFO, 0x01);
+	qmi_param_append(param, PARAM_REPORT_SIGNAL_STRENGTH, sizeof(ss), &ss);
+	qmi_param_append_uint8(param, PARAM_REPORT_RF_INFO, 0x01);
 
-	if (qmi_service_send(data->nas, QMI_NAS_SET_EVENT, param,
-					set_event_cb, netreg, NULL) > 0)
+	if (qmi_service_send(data->nas, QMI_NAS_SET_EVENT_REPORT, param,
+					set_event_report_cb, netreg, NULL) > 0)
 		return;
 
 	qmi_param_free(param);
-
-done:
-	ofono_netreg_register(netreg);
+error:
+	ofono_netreg_remove(netreg);
 }
 
 static int qmi_netreg_probe(struct ofono_netreg *netreg,
@@ -569,7 +681,7 @@ static int qmi_netreg_probe(struct ofono_netreg *netreg,
 
 	DBG("");
 
-	data = g_new0(struct netreg_data, 1);
+	data = l_new(struct netreg_data, 1);
 
 	data->operator.name[0] = '\0';
 	data->operator.mcc[0] = '\0';
@@ -578,6 +690,9 @@ static int qmi_netreg_probe(struct ofono_netreg *netreg,
 	data->operator.tech = -1;
 
 	data->current_rat = QMI_NAS_NETWORK_RAT_NO_CHANGE;
+	data->is_roaming = false;
+	data->lac = -1;
+	data->cellid = -1;
 
 	ofono_netreg_set_data(netreg, data);
 
@@ -595,15 +710,35 @@ static void qmi_netreg_remove(struct ofono_netreg *netreg)
 
 	ofono_netreg_set_data(netreg, NULL);
 
-	qmi_service_unregister_all(data->nas);
+	if (data->event_indication_id) {
+		qmi_service_unregister(data->nas, data->event_indication_id);
+		data->event_indication_id = 0;
+	}
 
-	qmi_service_unref(data->nas);
+	if (data->serving_system_indication_id) {
+		qmi_service_unregister(data->nas,
+					data->serving_system_indication_id);
+		data->serving_system_indication_id = 0;
+	}
 
-	g_free(data);
+	if (data->system_info_indication_id) {
+		qmi_service_unregister(data->nas,
+					data->system_info_indication_id);
+		data->system_info_indication_id = 0;
+	}
+
+	if (data->signal_info_indication_id) {
+		qmi_service_unregister(data->nas,
+					data->signal_info_indication_id);
+		data->signal_info_indication_id = 0;
+	}
+
+	qmi_service_free(data->nas);
+
+	l_free(data);
 }
 
-static struct ofono_netreg_driver driver = {
-	.name			= "qmimodem",
+static const struct ofono_netreg_driver driver = {
 	.probe			= qmi_netreg_probe,
 	.remove			= qmi_netreg_remove,
 	.registration_status	= qmi_registration_status,
@@ -614,12 +749,4 @@ static struct ofono_netreg_driver driver = {
 	.strength		= qmi_signal_strength,
 };
 
-void qmi_netreg_init(void)
-{
-	ofono_netreg_driver_register(&driver);
-}
-
-void qmi_netreg_exit(void)
-{
-	ofono_netreg_driver_unregister(&driver);
-}
+OFONO_ATOM_DRIVER_BUILTIN(netreg, qmimodem, &driver)
