@@ -101,9 +101,7 @@ static int gobi_probe(struct ofono_modem *modem)
 
 	DBG("%p", modem);
 
-	data = g_try_new0(struct gobi_data, 1);
-	if (!data)
-		return -ENOMEM;
+	data = l_new(struct gobi_data, 1);
 
 	kernel_driver = ofono_modem_get_string(modem, "KernelDriver");
 	DBG("kernel_driver: %s", kernel_driver);
@@ -119,16 +117,17 @@ static int gobi_probe(struct ofono_modem *modem)
 	DBG("net: %s (%d)", data->main_net_name, data->main_net_ifindex);
 
 	ofono_modem_set_data(modem, data);
+	ofono_modem_set_capabilities(modem, OFONO_MODEM_CAPABILITY_LTE);
 
 	return 0;
 }
 
 static void cleanup_services(struct gobi_data *data)
 {
-	qmi_service_unref(data->dms);
+	qmi_service_free(data->dms);
 	data->dms = NULL;
 
-	qmi_service_unref(data->wda);
+	qmi_service_free(data->wda);
 	data->wda = NULL;
 }
 
@@ -149,7 +148,7 @@ static void gobi_remove(struct ofono_modem *modem)
 
 	qmi_device_free(data->device);
 
-	g_free(data);
+	l_free(data);
 }
 
 static void shutdown_cb(void *user_data)
@@ -175,7 +174,8 @@ static void shutdown_device(struct ofono_modem *modem)
 
 	cleanup_services(data);
 
-	qmi_device_shutdown(data->device, shutdown_cb, modem, NULL);
+	if (qmi_device_shutdown(data->device, shutdown_cb, modem, NULL) < 0)
+		shutdown_cb(modem);
 }
 
 static void power_reset_cb(struct qmi_result *result, void *user_data)
@@ -330,7 +330,7 @@ static void create_wda_cb(struct qmi_service *service, void *user_data)
 		goto error;
 	}
 
-	data->wda = qmi_service_ref(service);
+	data->wda = service;
 
 	if (qmi_service_send(data->wda, QMI_WDA_GET_DATA_FORMAT, NULL,
 				get_data_format_cb, modem, NULL) > 0)
@@ -354,7 +354,7 @@ static void create_dms_cb(struct qmi_service *service, void *user_data)
 	if (!service)
 		goto error;
 
-	data->dms = qmi_service_ref(service);
+	data->dms = service;
 
 	if (qmi_service_create(data->device, QMI_SERVICE_WDA,
 					create_wda_cb, modem, NULL))
@@ -421,16 +421,27 @@ static void discover_cb(void *user_data)
 static int gobi_enable(struct ofono_modem *modem)
 {
 	struct gobi_data *data = ofono_modem_get_data(modem);
-	const char *device;
+	const char *kernel_driver;
 	int r;
 
 	DBG("%p", modem);
 
-	device = ofono_modem_get_string(modem, "Device");
-	if (!device)
+	kernel_driver = ofono_modem_get_string(modem, "KernelDriver");
+	if (!kernel_driver)
 		return -EINVAL;
 
-	data->device = qmi_device_new_qmux(device);
+	if (!strcmp(kernel_driver, "qrtr"))
+		data->device = qmi_device_new_qrtr();
+	else {
+		const char *device;
+
+		device = ofono_modem_get_string(modem, "Device");
+		if (!device)
+			return -EINVAL;
+
+		data->device = qmi_device_new_qmux(device);
+	}
+
 	if (!data->device)
 		return -EIO;
 
@@ -460,9 +471,6 @@ static int gobi_disable(struct ofono_modem *modem)
 
 	DBG("%p", modem);
 
-	qmi_service_cancel_all(data->dms);
-	qmi_service_unregister_all(data->dms);
-
 	/*
 	 * Telit QMI modem must remain online. If powered down, it also
 	 * powers down the sim card, and QMI interface has no way to bring
