@@ -48,6 +48,8 @@ static const char *none_prefix[] = { NULL };
 
 struct sierra_data {
 	GAtChat *modem;
+	gboolean have_sim;
+	struct at_util_sim_state_query *sim_state_query;
 };
 
 static void sierra_debug(const char *str, void *user_data)
@@ -80,6 +82,9 @@ static void sierra_remove(struct ofono_modem *modem)
 
 	ofono_modem_set_data(modem, NULL);
 
+	/* Cleanup potential SIM state polling */
+	at_util_sim_state_query_free(data->sim_state_query);
+
 	/* Cleanup after hot-unplug */
 	g_at_chat_unref(data->modem);
 
@@ -89,34 +94,22 @@ static void sierra_remove(struct ofono_modem *modem)
 static GAtChat *open_device(struct ofono_modem *modem,
 				const char *key, char *debug)
 {
-	const char *device;
-	GAtSyntax *syntax;
-	GIOChannel *channel;
-	GAtChat *chat;
+	return at_util_open_device(modem, key, sierra_debug, debug, NULL);
+}
 
-	device = ofono_modem_get_string(modem, key);
-	if (device == NULL)
-		return NULL;
+static void sim_state_cb(gboolean present, gpointer user_data)
+{
+	struct ofono_modem *modem = user_data;
+	struct sierra_data *data = ofono_modem_get_data(modem);
 
-	DBG("%s %s", key, device);
+	DBG("%p", modem);
 
-	channel = g_at_tty_open(device, NULL);
-	if (channel == NULL)
-		return NULL;
+	at_util_sim_state_query_free(data->sim_state_query);
+	data->sim_state_query = NULL;
 
-	syntax = g_at_syntax_new_gsm_permissive();
-	chat = g_at_chat_new(channel, syntax);
-	g_at_syntax_unref(syntax);
+	data->have_sim = present;
+	ofono_modem_set_powered(modem, TRUE);
 
-	g_io_channel_unref(channel);
-
-	if (chat == NULL)
-		return NULL;
-
-	if (getenv("OFONO_AT_DEBUG"))
-		g_at_chat_set_debug(chat, sierra_debug, debug);
-
-	return chat;
 }
 
 static void cfun_enable(gboolean ok, GAtResult *result, gpointer user_data)
@@ -129,9 +122,13 @@ static void cfun_enable(gboolean ok, GAtResult *result, gpointer user_data)
 	if (!ok) {
 		g_at_chat_unref(data->modem);
 		data->modem = NULL;
+		ofono_modem_set_powered(modem, FALSE);
+		return;
 	}
 
-	ofono_modem_set_powered(modem, ok);
+	data->sim_state_query = at_util_sim_state_query_new(data->modem,
+						2, 20, sim_state_cb, modem,
+						NULL);
 }
 
 static int sierra_enable(struct ofono_modem *modem)
@@ -222,7 +219,7 @@ static void sierra_pre_sim(struct ofono_modem *modem)
 	sim = ofono_sim_create(modem, OFONO_VENDOR_SIERRA,
 					"atmodem", data->modem);
 
-	if (sim)
+	if (sim && data->have_sim == TRUE)
 		ofono_sim_inserted_notify(sim, TRUE);
 }
 
@@ -253,7 +250,6 @@ static void sierra_post_online(struct ofono_modem *modem)
 }
 
 static struct ofono_modem_driver sierra_driver = {
-	.name		= "sierra",
 	.probe		= sierra_probe,
 	.remove		= sierra_remove,
 	.enable		= sierra_enable,
@@ -264,15 +260,4 @@ static struct ofono_modem_driver sierra_driver = {
 	.post_online	= sierra_post_online,
 };
 
-static int sierra_init(void)
-{
-	return ofono_modem_driver_register(&sierra_driver);
-}
-
-static void sierra_exit(void)
-{
-	ofono_modem_driver_unregister(&sierra_driver);
-}
-
-OFONO_PLUGIN_DEFINE(sierra, "Sierra Wireless modem driver", VERSION,
-			OFONO_PLUGIN_PRIORITY_DEFAULT, sierra_init, sierra_exit)
+OFONO_MODEM_DRIVER_BUILTIN(sierra, &sierra_driver)
