@@ -39,8 +39,6 @@
 #define DBUS_TYPE_UNIX_FD -1
 #endif
 
-static GSList *g_drivers = NULL;
-
 struct ofono_location_reporting {
 	DBusMessage *pending;
 	const struct ofono_location_reporting_driver *driver;
@@ -71,7 +69,7 @@ static DBusMessage *location_reporting_get_properties(DBusConnection *conn,
 	DBusMessageIter iter;
 	DBusMessageIter dict;
 	const char *type;
-	int value;
+	dbus_bool_t value;
 
 	reply = dbus_message_new_method_return(msg);
 	if (reply == NULL)
@@ -102,14 +100,14 @@ static void client_remove(struct ofono_location_reporting *lr)
 		lr->disconnect_watch = 0;
 	}
 
-	g_free(lr->client_owner);
+	l_free(lr->client_owner);
 }
 
 static void signal_enabled(const struct ofono_location_reporting *lr)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
 	const char *path = __ofono_atom_get_path(lr->atom);
-	int value = lr->enabled;
+	dbus_bool_t value = lr->enabled;
 
 	ofono_dbus_signal_property_changed(conn, path,
 					OFONO_LOCATION_REPORTING_INTERFACE,
@@ -123,7 +121,6 @@ static void client_exited_disable_cb(const struct ofono_error *error,
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
 		ofono_error("Disabling location-reporting failed");
-
 		return;
 	}
 
@@ -139,7 +136,7 @@ static void client_exited(DBusConnection *conn, void *data)
 
 	lr->disconnect_watch = 0;
 
-	lr->driver->disable(lr, client_exited_disable_cb , lr);
+	lr->driver->disable(lr, client_exited_disable_cb, lr);
 }
 
 static void location_reporting_disable_cb(const struct ofono_error *error,
@@ -153,7 +150,6 @@ static void location_reporting_disable_cb(const struct ofono_error *error,
 
 		reply = __ofono_error_failed(lr->pending);
 		__ofono_dbus_pending_reply(&lr->pending, reply);
-
 		return;
 	}
 
@@ -178,12 +174,11 @@ static void location_reporting_enable_cb(const struct ofono_error *error,
 
 		reply = __ofono_error_failed(lr->pending);
 		__ofono_dbus_pending_reply(&lr->pending, reply);
-
 		return;
 	}
 
 	lr->enabled = TRUE;
-	lr->client_owner = g_strdup(dbus_message_get_sender(lr->pending));
+	lr->client_owner = l_strdup(dbus_message_get_sender(lr->pending));
 	lr->disconnect_watch = g_dbus_add_disconnect_watch(conn,
 				lr->client_owner, client_exited, lr, NULL);
 
@@ -220,7 +215,6 @@ static DBusMessage *location_reporting_release(DBusConnection *conn,
 	struct ofono_location_reporting *lr = data;
 	const char *caller = dbus_message_get_sender(msg);
 
-
 	/*
 	 * Avoid a race by not trying to release the device if there is a
 	 * pending message or client already signaled it's exiting. In the
@@ -243,44 +237,24 @@ static DBusMessage *location_reporting_release(DBusConnection *conn,
 	return NULL;
 }
 
-static GDBusMethodTable location_reporting_methods[] = {
-	{ "GetProperties",  "",    "a{sv}", location_reporting_get_properties },
-	{ "Request",        "",    "h",     location_reporting_request,
-						G_DBUS_METHOD_FLAG_ASYNC },
-	{ "Release",        "",    "",      location_reporting_release,
-						G_DBUS_METHOD_FLAG_ASYNC },
+static const GDBusMethodTable location_reporting_methods[] = {
+	{ GDBUS_METHOD("GetProperties",
+			NULL, GDBUS_ARGS({ "properties", "a{sv}" }),
+			location_reporting_get_properties) },
+	{ GDBUS_ASYNC_METHOD("Request",
+			NULL, GDBUS_ARGS({ "fd", "h" }),
+			location_reporting_request) },
+	{ GDBUS_ASYNC_METHOD("Release", NULL, NULL,
+					location_reporting_release) },
 	{ }
 };
 
-static GDBusSignalTable location_reporting_signals[] = {
-	{ "PropertyChanged",	"sv" },
+static const GDBusSignalTable location_reporting_signals[] = {
+	{ GDBUS_SIGNAL("PropertyChanged",
+			GDBUS_ARGS({ "name", "s" }, { "value", "v" })) },
 	{ }
 };
 
-int ofono_location_reporting_driver_register(
-				const struct ofono_location_reporting_driver *d)
-{
-	DBG("driver: %p, name: %s", d, d->name);
-
-	if (d == NULL || d->probe == NULL)
-		return -EINVAL;
-
-	g_drivers = g_slist_prepend(g_drivers, (void *) d);
-
-	return 0;
-}
-
-void ofono_location_reporting_driver_unregister(
-				const struct ofono_location_reporting_driver *d)
-{
-	DBG("driver: %p, name: %s", d, d->name);
-
-	if (d == NULL)
-		return;
-
-	g_drivers = g_slist_remove(g_drivers, (void *) d);
-}
-
 struct ofono_modem *ofono_location_reporting_get_modem(
 					struct ofono_location_reporting *lr)
 {
@@ -314,44 +288,8 @@ static void location_reporting_remove(struct ofono_atom *atom)
 	g_free(lr);
 }
 
-struct ofono_location_reporting *ofono_location_reporting_create(
-						struct ofono_modem *modem,
-						unsigned int vendor,
-						const char *driver, void *data)
-{
-	struct ofono_location_reporting *lr;
-	GSList *l;
-
-	if (driver == NULL)
-		return NULL;
-
-	/* Only D-Bus >= 1.3 supports fd-passing */
-	if (DBUS_TYPE_UNIX_FD == -1)
-		return NULL;
-
-	lr = g_try_new0(struct ofono_location_reporting, 1);
-	if (lr == NULL)
-		return NULL;
-
-	lr->atom = __ofono_modem_add_atom(modem,
-					OFONO_ATOM_TYPE_LOCATION_REPORTING,
-					location_reporting_remove, lr);
-
-	for (l = g_drivers; l; l = l->next) {
-		const struct ofono_location_reporting_driver *drv = l->data;
-
-		if (g_strcmp0(drv->name, driver) != 0)
-			continue;
-
-		if (drv->probe(lr, vendor, data) < 0)
-			continue;
-
-		lr->driver = drv;
-		break;
-	}
-
-	return lr;
-}
+OFONO_DEFINE_ATOM_CREATE(location_reporting,
+				OFONO_ATOM_TYPE_LOCATION_REPORTING)
 
 void ofono_location_reporting_register(struct ofono_location_reporting *lr)
 {
@@ -366,7 +304,6 @@ void ofono_location_reporting_register(struct ofono_location_reporting *lr)
 					NULL, lr, NULL)) {
 		ofono_error("Could not create %s interface",
 					OFONO_LOCATION_REPORTING_INTERFACE);
-
 		return;
 	}
 
