@@ -36,7 +36,6 @@
 
 #define DEFAULT_POWERED_TIMEOUT (20)
 
-static GSList *g_driver_list;
 static GSList *g_modem_list;
 
 static int next_modem_id;
@@ -89,6 +88,7 @@ struct ofono_modem {
 	void			*driver_data;
 	char			*driver_type;
 	char			*name;
+	unsigned int		capabilities;
 };
 
 struct ofono_devinfo {
@@ -138,6 +138,48 @@ static const char *modem_type_to_string(enum ofono_modem_type type)
 	return "unknown";
 }
 
+static char **get_features(struct ofono_modem *modem)
+{
+	char **features = l_new(char *, g_slist_length(modem->feature_list) + 1);
+	unsigned int i;
+	GSList *l;
+
+	for (i = 0, l = modem->feature_list; l; l = l->next, i++)
+		features[i] = l->data;
+
+	return features;
+}
+
+static char **get_interfaces(struct ofono_modem *modem)
+{
+	char **interfaces = l_new(char *,
+				g_slist_length(modem->interface_list) + 1);
+	unsigned int i;
+	GSList *l;
+
+	for (i = 0, l = modem->interface_list; l; l = l->next, i++)
+		interfaces[i] = l->data;
+
+	return interfaces;
+}
+
+static char **get_capabilities(struct ofono_modem *modem)
+{
+	char **capabilities;
+	unsigned int i = 0;
+
+	if (!modem->capabilities)
+		return NULL;
+
+	capabilities = l_new(char *,
+			__builtin_popcount(modem->capabilities) + 1);
+
+	if (modem->capabilities & OFONO_MODEM_CAPABILITY_LTE)
+		capabilities[i++] = "lte";
+
+	return capabilities;
+}
+
 unsigned int __ofono_modem_callid_next(struct ofono_modem *modem)
 {
 	unsigned int i;
@@ -438,13 +480,6 @@ static void flush_atoms(struct ofono_modem *modem, enum modem_state new_state)
 			continue;
 		}
 
-		__ofono_atom_unregister(atom);
-
-		if (atom->destruct)
-			atom->destruct(atom);
-
-		g_free(atom);
-
 		if (prev)
 			prev->next = cur->next;
 		else
@@ -453,14 +488,21 @@ static void flush_atoms(struct ofono_modem *modem, enum modem_state new_state)
 		tmp = cur;
 		cur = cur->next;
 		g_slist_free_1(tmp);
+
+		__ofono_atom_unregister(atom);
+
+		if (atom->destruct)
+			atom->destruct(atom);
+
+		g_free(atom);
 	}
 }
 
-const void *__ofono_atom_driver_builtin_find(const char *name,
-				const struct ofono_atom_driver_desc *start,
-				const struct ofono_atom_driver_desc *stop)
+const void *__ofono_driver_builtin_find(const char *name,
+				const struct ofono_driver_desc *start,
+				const struct ofono_driver_desc *stop)
 {
-	const struct ofono_atom_driver_desc *desc;
+	const struct ofono_driver_desc *desc;
 
 	if (!name)
 		return NULL;
@@ -821,12 +863,11 @@ void __ofono_modem_append_properties(struct ofono_modem *modem,
 {
 	char **interfaces;
 	char **features;
-	int i;
-	GSList *l;
 	struct ofono_devinfo *info;
 	dbus_bool_t emergency = ofono_modem_get_emergency_mode(modem);
 	const char *strtype;
 	const char *system_path;
+	char **capabilities;
 
 	ofono_dbus_dict_append(dict, "Online", DBUS_TYPE_BOOLEAN,
 				&modem->online);
@@ -872,19 +913,15 @@ void __ofono_modem_append_properties(struct ofono_modem *modem,
 		ofono_dbus_dict_append(dict, "SystemPath", DBUS_TYPE_STRING,
 					&system_path);
 
-	interfaces = g_new0(char *, g_slist_length(modem->interface_list) + 1);
-	for (i = 0, l = modem->interface_list; l; l = l->next, i++)
-		interfaces[i] = l->data;
+	interfaces = get_interfaces(modem);
 	ofono_dbus_dict_append_array(dict, "Interfaces", DBUS_TYPE_STRING,
 					&interfaces);
-	g_free(interfaces);
+	l_free(interfaces);
 
-	features = g_new0(char *, g_slist_length(modem->feature_list) + 1);
-	for (i = 0, l = modem->feature_list; l; l = l->next, i++)
-		features[i] = l->data;
+	features = get_features(modem);
 	ofono_dbus_dict_append_array(dict, "Features", DBUS_TYPE_STRING,
 					&features);
-	g_free(features);
+	l_free(features);
 
 	if (modem->name)
 		ofono_dbus_dict_append(dict, "Name", DBUS_TYPE_STRING,
@@ -892,6 +929,13 @@ void __ofono_modem_append_properties(struct ofono_modem *modem,
 
 	strtype = modem_type_to_string(modem->driver->modem_type);
 	ofono_dbus_dict_append(dict, "Type", DBUS_TYPE_STRING, &strtype);
+
+	capabilities = get_capabilities(modem);
+	if (capabilities) {
+		ofono_dbus_dict_append_array(dict, "Capabilities",
+						DBUS_TYPE_STRING, &capabilities);
+		l_free(capabilities);
+	}
 }
 
 static DBusMessage *modem_get_properties(DBusConnection *conn,
@@ -960,7 +1004,7 @@ static void lockdown_remove(struct ofono_modem *modem)
 		modem->lock_watch = 0;
 	}
 
-	g_free(modem->lock_owner);
+	l_free(modem->lock_owner);
 	modem->lock_owner = NULL;
 
 	modem->lockdown = FALSE;
@@ -1053,14 +1097,14 @@ static DBusMessage *set_property_lockdown(struct ofono_modem *modem,
 	if (ofono_modem_get_emergency_mode(modem) == TRUE)
 		return __ofono_error_emergency_active(msg);
 
-	modem->lock_owner = g_strdup(caller);
+	modem->lock_owner = l_strdup(caller);
 
 	modem->lock_watch = g_dbus_add_disconnect_watch(conn,
 				modem->lock_owner, lockdown_disconnect,
 				modem, NULL);
 
 	if (modem->lock_watch == 0) {
-		g_free(modem->lock_owner);
+		l_free(modem->lock_owner);
 		modem->lock_owner = NULL;
 
 		return __ofono_error_failed(msg);
@@ -1219,6 +1263,19 @@ void ofono_modem_set_powered(struct ofono_modem *modem, ofono_bool_t powered)
 		modem->timeout = 0;
 	}
 
+	if (powered && powered != modem->powered) {
+		char **capabilities = get_capabilities(modem);
+
+		if (capabilities) {
+			ofono_dbus_signal_array_property_changed(conn,
+						modem->path,
+						OFONO_MODEM_INTERFACE,
+						"Capabilities", DBUS_TYPE_STRING,
+						&capabilities);
+			l_free(capabilities);
+		}
+	}
+
 	if (modem->powered_pending != modem->powered &&
 						modem->pending != NULL) {
 		DBusMessage *reply;
@@ -1292,26 +1349,20 @@ static gboolean trigger_interface_update(void *data)
 	DBusConnection *conn = ofono_dbus_get_connection();
 	char **interfaces;
 	char **features;
-	GSList *l;
-	int i;
 
-	interfaces = g_new0(char *, g_slist_length(modem->interface_list) + 1);
-	for (i = 0, l = modem->interface_list; l; l = l->next, i++)
-		interfaces[i] = l->data;
+	interfaces = get_interfaces(modem);
 	ofono_dbus_signal_array_property_changed(conn, modem->path,
 						OFONO_MODEM_INTERFACE,
 						"Interfaces", DBUS_TYPE_STRING,
 						&interfaces);
-	g_free(interfaces);
+	l_free(interfaces);
 
-	features = g_new0(char *, g_slist_length(modem->feature_list) + 1);
-	for (i = 0, l = modem->feature_list; l; l = l->next, i++)
-		features[i] = l->data;
+	features = get_features(modem);
 	ofono_dbus_signal_array_property_changed(conn, modem->path,
 						OFONO_MODEM_INTERFACE,
 						"Features", DBUS_TYPE_STRING,
 						&features);
-	g_free(features);
+	l_free(features);
 
 	modem->interface_update = 0;
 
@@ -1353,12 +1404,12 @@ void ofono_modem_add_interface(struct ofono_modem *modem,
 	const char *feature;
 
 	modem->interface_list = g_slist_prepend(modem->interface_list,
-						g_strdup(interface));
+						l_strdup(interface));
 
 	feature = get_feature(interface);
 	if (feature)
 		modem->feature_list = g_slist_prepend(modem->feature_list,
-							g_strdup(feature));
+							l_strdup(feature));
 
 	if (modem->interface_update != 0)
 		return;
@@ -1380,7 +1431,7 @@ void ofono_modem_remove_interface(struct ofono_modem *modem,
 		return;
 	}
 
-	g_free(found->data);
+	l_free(found->data);
 	modem->interface_list = g_slist_remove(modem->interface_list,
 						found->data);
 
@@ -1389,7 +1440,7 @@ void ofono_modem_remove_interface(struct ofono_modem *modem,
 		found = g_slist_find_custom(modem->feature_list, feature,
 						(GCompareFunc) strcmp);
 		if (found) {
-			g_free(found->data);
+			l_free(found->data);
 			modem->feature_list =
 				g_slist_remove(modem->feature_list,
 						found->data);
@@ -1412,7 +1463,7 @@ static void query_svn_cb(const struct ofono_error *error,
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
 		return;
 
-	info->svn = g_strdup(svn);
+	info->svn = l_strdup(svn);
 
 	ofono_dbus_signal_property_changed(conn, path, OFONO_MODEM_INTERFACE,
 			"SoftwareVersionNumber", DBUS_TYPE_STRING, &info->svn);
@@ -1436,7 +1487,7 @@ static void query_serial_cb(const struct ofono_error *error,
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
 		goto out;
 
-	info->serial = g_strdup(serial);
+	info->serial = l_strdup(serial);
 
 	ofono_dbus_signal_property_changed(conn, path,
 						OFONO_MODEM_INTERFACE,
@@ -1464,7 +1515,7 @@ static void query_revision_cb(const struct ofono_error *error,
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
 		goto out;
 
-	info->revision = g_strdup(revision);
+	info->revision = l_strdup(revision);
 
 	ofono_dbus_signal_property_changed(conn, path,
 						OFONO_MODEM_INTERFACE,
@@ -1495,7 +1546,7 @@ static void query_model_cb(const struct ofono_error *error,
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
 		goto out;
 
-	info->model = g_strdup(model);
+	info->model = l_strdup(model);
 
 	ofono_dbus_signal_property_changed(conn, path,
 						OFONO_MODEM_INTERFACE,
@@ -1527,7 +1578,7 @@ static void query_manufacturer_cb(const struct ofono_error *error,
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
 		goto out;
 
-	info->manufacturer = g_strdup(manufacturer);
+	info->manufacturer = l_strdup(manufacturer);
 
 	ofono_dbus_signal_property_changed(conn, path,
 						OFONO_MODEM_INTERFACE,
@@ -1647,19 +1698,19 @@ static void devinfo_unregister(struct ofono_atom *atom)
 {
 	struct ofono_devinfo *info = __ofono_atom_get_data(atom);
 
-	g_free(info->manufacturer);
+	l_free(info->manufacturer);
 	info->manufacturer = NULL;
 
-	g_free(info->model);
+	l_free(info->model);
 	info->model = NULL;
 
-	g_free(info->revision);
+	l_free(info->revision);
 	info->revision = NULL;
 
-	g_free(info->serial);
+	l_free(info->serial);
 	info->serial = NULL;
 
-	g_free(info->svn);
+	l_free(info->svn);
 	info->svn = NULL;
 }
 
@@ -1726,7 +1777,7 @@ static int set_modem_property(struct ofono_modem *modem, const char *name,
 
 	switch (type) {
 	case PROPERTY_TYPE_STRING:
-		property->value = g_strdup((const char *) value);
+		property->value = l_strdup((const char *) value);
 		break;
 	case PROPERTY_TYPE_INTEGER:
 		property->value = g_memdup2(value, sizeof(int));
@@ -1738,7 +1789,7 @@ static int set_modem_property(struct ofono_modem *modem, const char *name,
 		break;
 	}
 
-	g_hash_table_replace(modem->properties, g_strdup(name), property);
+	g_hash_table_replace(modem->properties, l_strdup(name), property);
 
 	return 0;
 }
@@ -1834,9 +1885,9 @@ void ofono_modem_set_powered_timeout_hint(struct ofono_modem *modem,
 void ofono_modem_set_name(struct ofono_modem *modem, const char *name)
 {
 	if (modem->name)
-		g_free(modem->name);
+		l_free(modem->name);
 
-	modem->name = g_strdup(name);
+	modem->name = l_strdup(name);
 
 	if (modem->driver) {
 		DBusConnection *conn = ofono_dbus_get_connection();
@@ -1858,8 +1909,19 @@ void ofono_modem_set_driver(struct ofono_modem *modem, const char *type)
 	if (strlen(type) > 16)
 		return;
 
-	g_free(modem->driver_type);
-	modem->driver_type = g_strdup(type);
+	l_free(modem->driver_type);
+	modem->driver_type = l_strdup(type);
+}
+
+void ofono_modem_set_capabilities(struct ofono_modem *modem,
+						unsigned int capabilities)
+{
+	if (!modem)
+		return;
+
+	DBG("%u", capabilities);
+
+	modem->capabilities = capabilities;
 }
 
 struct ofono_modem *ofono_modem_create(const char *name, const char *type)
@@ -1888,10 +1950,10 @@ struct ofono_modem *ofono_modem_create(const char *name, const char *type)
 	if (modem == NULL)
 		return modem;
 
-	modem->path = g_strdup(path);
-	modem->driver_type = g_strdup(type);
+	modem->path = l_strdup(path);
+	modem->driver_type = l_strdup(type);
 	modem->properties = g_hash_table_new_full(g_str_hash, g_str_equal,
-						g_free, unregister_property);
+						l_free, unregister_property);
 	modem->timeout_hint = DEFAULT_POWERED_TIMEOUT;
 
 	g_modem_list = g_slist_prepend(g_modem_list, modem);
@@ -1918,12 +1980,13 @@ static void sim_watch(struct ofono_atom *atom,
 							modem, NULL);
 }
 
-void __ofono_modemwatch_init(void)
+static int modemwatch_init(void)
 {
 	g_modemwatches = __ofono_watchlist_new(g_free);
+	return 0;
 }
 
-void __ofono_modemwatch_cleanup(void)
+static void modemwatch_cleanup(void)
 {
 	__ofono_watchlist_free(g_modemwatches);
 }
@@ -2006,10 +2069,14 @@ ofono_bool_t ofono_modem_is_registered(struct ofono_modem *modem)
 	return TRUE;
 }
 
+extern struct ofono_driver_desc __start___modem[];
+extern struct ofono_driver_desc __stop___modem[];
+
 int ofono_modem_register(struct ofono_modem *modem)
 {
+	const struct ofono_modem_driver *drv;
 	DBusConnection *conn = ofono_dbus_get_connection();
-	GSList *l;
+	int r;
 
 	DBG("%p", modem);
 
@@ -2022,22 +2089,18 @@ int ofono_modem_register(struct ofono_modem *modem)
 	if (modem->driver != NULL)
 		return -EALREADY;
 
-	for (l = g_driver_list; l; l = l->next) {
-		const struct ofono_modem_driver *drv = l->data;
-
-		if (g_strcmp0(drv->name, modem->driver_type))
-			continue;
-
-		if (drv->probe(modem) < 0)
-			continue;
-
-		modem->driver = drv;
-		break;
-	}
-
-	if (modem->driver == NULL)
+	drv = __ofono_driver_builtin_find(modem->driver_type,
+						__start___modem,
+						__stop___modem);
+	if (!drv || !drv->probe)
 		return -ENODEV;
 
+	r = drv->probe(modem);
+	if (r < 0)
+		return r;
+
+	modem->driver = drv;
+
 	if (!g_dbus_register_interface(conn, modem->path,
 					OFONO_MODEM_INTERFACE,
 					modem_methods, modem_signals, NULL,
@@ -2052,7 +2115,7 @@ int ofono_modem_register(struct ofono_modem *modem)
 		return -EIO;
 	}
 
-	g_free(modem->driver_type);
+	l_free(modem->driver_type);
 	modem->driver_type = NULL;
 
 	modem->atom_watches = __ofono_watchlist_new(g_free);
@@ -2102,10 +2165,10 @@ static void modem_unregister(struct ofono_modem *modem)
 	modem->sim_watch = 0;
 	modem->sim_ready_watch = 0;
 
-	g_slist_free_full(modem->interface_list, g_free);
+	g_slist_free_full(modem->interface_list, l_free);
 	modem->interface_list = NULL;
 
-	g_slist_free_full(modem->feature_list, g_free);
+	g_slist_free_full(modem->feature_list, l_free);
 	modem->feature_list = NULL;
 
 	if (modem->timeout) {
@@ -2158,9 +2221,9 @@ void ofono_modem_remove(struct ofono_modem *modem)
 
 	g_modem_list = g_slist_remove(g_modem_list, modem);
 
-	g_free(modem->driver_type);
-	g_free(modem->name);
-	g_free(modem->path);
+	l_free(modem->driver_type);
+	l_free(modem->name);
+	l_free(modem->path);
 	g_free(modem);
 }
 
@@ -2197,37 +2260,6 @@ void __ofono_modem_sim_reset(struct ofono_modem *modem)
 	modem_change_state(modem, MODEM_STATE_PRE_SIM);
 }
 
-int ofono_modem_driver_register(const struct ofono_modem_driver *d)
-{
-	DBG("driver: %p, name: %s", d, d->name);
-
-	if (d->probe == NULL)
-		return -EINVAL;
-
-	g_driver_list = g_slist_prepend(g_driver_list, (void *) d);
-
-	return 0;
-}
-
-void ofono_modem_driver_unregister(const struct ofono_modem_driver *d)
-{
-	GSList *l;
-	struct ofono_modem *modem;
-
-	DBG("driver: %p, name: %s", d, d->name);
-
-	g_driver_list = g_slist_remove(g_driver_list, (void *) d);
-
-	for (l = g_modem_list; l; l = l->next) {
-		modem = l->data;
-
-		if (modem->driver != d)
-			continue;
-
-		modem_unregister(modem);
-	}
-}
-
 void __ofono_modem_shutdown(void)
 {
 	struct ofono_modem *modem;
@@ -2319,3 +2351,5 @@ void __ofono_modem_dec_emergency_mode(struct ofono_modem *modem)
 out:
 	modem->emergency--;
 }
+
+OFONO_MODULE(modemwatch, modemwatch_init, modemwatch_cleanup)
