@@ -33,29 +33,32 @@
 
 #include "ofono.h"
 #include "common.h"
+#include "storage.h"
 
+#define SETTINGS_STORE "radiosetting"
+#define SETTINGS_GROUP "Settings"
 #define RADIO_SETTINGS_FLAG_CACHED 0x1
 
-static GSList *g_drivers = NULL;
-
 struct ofono_radio_settings {
 	DBusMessage *pending;
 	int flags;
-	enum ofono_radio_access_mode mode;
+	unsigned int mode;
 	enum ofono_radio_band_gsm band_gsm;
 	enum ofono_radio_band_umts band_umts;
 	ofono_bool_t fast_dormancy;
-	enum ofono_radio_access_mode pending_mode;
+	unsigned int pending_mode;
 	enum ofono_radio_band_gsm pending_band_gsm;
 	enum ofono_radio_band_umts pending_band_umts;
 	ofono_bool_t fast_dormancy_pending;
 	uint32_t available_rats;
+	GKeyFile *settings;
+	char *imsi;
 	const struct ofono_radio_settings_driver *driver;
 	void *driver_data;
 	struct ofono_atom *atom;
 };
 
-static const char *radio_access_mode_to_string(enum ofono_radio_access_mode m)
+static const char *radio_access_mode_to_string(unsigned int m)
 {
 	switch (m) {
 	case OFONO_RADIO_ACCESS_MODE_ANY:
@@ -66,13 +69,22 @@ static const char *radio_access_mode_to_string(enum ofono_radio_access_mode m)
 		return "umts";
 	case OFONO_RADIO_ACCESS_MODE_LTE:
 		return "lte";
-	default:
-		return "";
 	}
+
+	if (m == (OFONO_RADIO_ACCESS_MODE_UMTS|OFONO_RADIO_ACCESS_MODE_GSM))
+		return "umts,gsm";
+
+	if (m == (OFONO_RADIO_ACCESS_MODE_LTE|OFONO_RADIO_ACCESS_MODE_UMTS))
+		return "lte,umts";
+
+	if (m == (OFONO_RADIO_ACCESS_MODE_LTE|OFONO_RADIO_ACCESS_MODE_GSM))
+		return "lte,gsm";
+
+	return NULL;
 }
 
 static gboolean radio_access_mode_from_string(const char *str,
-					enum ofono_radio_access_mode *mode)
+					unsigned int *mode)
 
 {
 	if (g_str_equal(str, "any")) {
@@ -87,6 +99,15 @@ static gboolean radio_access_mode_from_string(const char *str,
 	} else if (g_str_equal(str, "lte")) {
 		*mode = OFONO_RADIO_ACCESS_MODE_LTE;
 		return TRUE;
+	} else if (g_str_equal(str, "umts,gsm")) {
+		*mode = OFONO_RADIO_ACCESS_MODE_UMTS|OFONO_RADIO_ACCESS_MODE_GSM;
+		return TRUE;
+	} else if (g_str_equal(str, "lte,umts")) {
+		*mode = OFONO_RADIO_ACCESS_MODE_LTE|OFONO_RADIO_ACCESS_MODE_UMTS;
+		return TRUE;
+	} else if (g_str_equal(str, "lte,gsm")) {
+		*mode = OFONO_RADIO_ACCESS_MODE_LTE|OFONO_RADIO_ACCESS_MODE_GSM;
+		return TRUE;
 	}
 
 	return FALSE;
@@ -109,7 +130,7 @@ static const char *radio_band_gsm_to_string(enum ofono_radio_band_gsm band)
 		return "1900";
 	}
 
-	return "";
+	return NULL;
 }
 
 static gboolean radio_band_gsm_from_string(const char *str,
@@ -155,14 +176,14 @@ static const char *radio_band_umts_to_string(enum ofono_radio_band_umts band)
 		return "2100";
 	}
 
-	return "";
+	return NULL;
 }
 
 static gboolean radio_band_umts_from_string(const char *str,
 					enum ofono_radio_band_umts *band)
 {
 	if (g_str_equal(str, "any")) {
-		*band = OFONO_RADIO_BAND_GSM_ANY;
+		*band = OFONO_RADIO_BAND_UMTS_ANY;
 		return TRUE;
 	} else if (g_str_equal(str, "850")) {
 		*band = OFONO_RADIO_BAND_UMTS_850;
@@ -306,6 +327,12 @@ static void radio_set_band(struct ofono_radio_settings *rs)
 						OFONO_RADIO_SETTINGS_INTERFACE,
 						"GsmBand", DBUS_TYPE_STRING,
 						&str_band);
+
+		if (rs->settings) {
+			g_key_file_set_integer(rs->settings, SETTINGS_GROUP,
+					"GsmBand", rs->band_gsm);
+			storage_sync(rs->imsi, SETTINGS_STORE, rs->settings);
+		}
 	}
 
 	if (rs->band_umts != rs->pending_band_umts) {
@@ -316,8 +343,13 @@ static void radio_set_band(struct ofono_radio_settings *rs)
 						OFONO_RADIO_SETTINGS_INTERFACE,
 						"UmtsBand", DBUS_TYPE_STRING,
 						&str_band);
-	}
 
+		if (rs->settings) {
+			g_key_file_set_integer(rs->settings, SETTINGS_GROUP,
+					"UmtsBand", rs->band_umts);
+			storage_sync(rs->imsi, SETTINGS_STORE, rs->settings);
+		}
+	}
 }
 
 static void radio_band_set_callback(const struct ofono_error *error,
@@ -345,7 +377,7 @@ static void radio_band_set_callback(const struct ofono_error *error,
 }
 
 static void radio_set_rat_mode(struct ofono_radio_settings *rs,
-				enum ofono_radio_access_mode mode)
+				unsigned int mode)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
 	const char *path;
@@ -363,6 +395,12 @@ static void radio_set_rat_mode(struct ofono_radio_settings *rs,
 						OFONO_RADIO_SETTINGS_INTERFACE,
 						"TechnologyPreference",
 						DBUS_TYPE_STRING, &str_mode);
+
+	if (rs->settings) {
+		g_key_file_set_integer(rs->settings, SETTINGS_GROUP,
+				"TechnologyPreference", rs->mode);
+		storage_sync(rs->imsi, SETTINGS_STORE, rs->settings);
+	}
 }
 
 static void radio_mode_set_callback(const struct ofono_error *error, void *data)
@@ -488,8 +526,7 @@ static void radio_query_band(struct ofono_radio_settings *rs)
 }
 
 static void radio_rat_mode_query_callback(const struct ofono_error *error,
-					enum ofono_radio_access_mode mode,
-					void *data)
+						int mode, void *data)
 {
 	struct ofono_radio_settings *rs = data;
 	DBusMessage *reply;
@@ -554,7 +591,7 @@ static DBusMessage *radio_set_property(DBusConnection *conn, DBusMessage *msg,
 
 	if (g_strcmp0(property, "TechnologyPreference") == 0) {
 		const char *value;
-		enum ofono_radio_access_mode mode;
+		unsigned int mode;
 
 		if (rs->driver->set_rat_mode == NULL)
 			return __ofono_error_not_implemented(msg);
@@ -573,7 +610,7 @@ static DBusMessage *radio_set_property(DBusConnection *conn, DBusMessage *msg,
 		rs->pending_mode = mode;
 
 		rs->driver->set_rat_mode(rs, mode, radio_mode_set_callback, rs);
-
+		/* will be saved in radiosettng on success response*/
 		return NULL;
 	} else if (g_strcmp0(property, "GsmBand") == 0) {
 		const char *value;
@@ -597,7 +634,7 @@ static DBusMessage *radio_set_property(DBusConnection *conn, DBusMessage *msg,
 
 		rs->driver->set_band(rs, band, rs->band_umts,
 					radio_band_set_callback, rs);
-
+		/* will be saved in radiosettng on success response*/
 		return NULL;
 	} else if (g_strcmp0(property, "UmtsBand") == 0) {
 		const char *value;
@@ -621,7 +658,7 @@ static DBusMessage *radio_set_property(DBusConnection *conn, DBusMessage *msg,
 
 		rs->driver->set_band(rs, rs->band_gsm, band,
 					radio_band_set_callback, rs);
-
+		/* will be saved in radiosettng on success response*/
 		return NULL;
 	} else if (g_strcmp0(property, "FastDormancy") == 0) {
 		dbus_bool_t value;
@@ -666,28 +703,6 @@ static const GDBusSignalTable radio_signals[] = {
 	{ }
 };
 
-int ofono_radio_settings_driver_register(const struct ofono_radio_settings_driver *d)
-{
-	DBG("driver: %p, name: %s", d, d->name);
-
-	if (d == NULL || d->probe == NULL)
-		return -EINVAL;
-
-	g_drivers = g_slist_prepend(g_drivers, (void *) d);
-
-	return 0;
-}
-
-void ofono_radio_settings_driver_unregister(const struct ofono_radio_settings_driver *d)
-{
-	DBG("driver: %p, name: %s", d, d->name);
-
-	if (d == NULL)
-		return;
-
-	g_drivers = g_slist_remove(g_drivers, (void *) d);
-}
-
 static void radio_settings_unregister(struct ofono_atom *atom)
 {
 	struct ofono_radio_settings *rs = __ofono_atom_get_data(atom);
@@ -697,6 +712,14 @@ static void radio_settings_unregister(struct ofono_atom *atom)
 
 	ofono_modem_remove_interface(modem, OFONO_RADIO_SETTINGS_INTERFACE);
 	g_dbus_unregister_interface(conn, path, OFONO_RADIO_SETTINGS_INTERFACE);
+
+	if (rs->settings) {
+		storage_close(rs->imsi, SETTINGS_STORE, rs->settings, TRUE);
+
+		l_free(rs->imsi);
+		rs->imsi = NULL;
+		rs->settings = NULL;
+	}
 }
 
 static void radio_settings_remove(struct ofono_atom *atom)
@@ -714,43 +737,11 @@ static void radio_settings_remove(struct ofono_atom *atom)
 	g_free(rs);
 }
 
-struct ofono_radio_settings *ofono_radio_settings_create(struct ofono_modem *modem,
-							unsigned int vendor,
-							const char *driver,
-							void *data)
-{
-	struct ofono_radio_settings *rs;
-	GSList *l;
+OFONO_DEFINE_ATOM_CREATE(radio_settings, OFONO_ATOM_TYPE_RADIO_SETTINGS, {
+	atom->mode = -1;
+})
 
-	if (driver == NULL)
-		return NULL;
-
-	rs = g_try_new0(struct ofono_radio_settings, 1);
-	if (rs == NULL)
-		return NULL;
-
-	rs->mode = -1;
-
-	rs->atom = __ofono_modem_add_atom(modem, OFONO_ATOM_TYPE_RADIO_SETTINGS,
-						radio_settings_remove, rs);
-
-	for (l = g_drivers; l; l = l->next) {
-		const struct ofono_radio_settings_driver *drv = l->data;
-
-		if (g_strcmp0(drv->name, driver) != 0)
-			continue;
-
-		if (drv->probe(rs, vendor, data) < 0)
-			continue;
-
-		rs->driver = drv;
-		break;
-	}
-
-	return rs;
-}
-
-void ofono_radio_settings_register(struct ofono_radio_settings *rs)
+static void ofono_radio_finish_register(struct ofono_radio_settings *rs)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
 	struct ofono_modem *modem = __ofono_atom_get_modem(rs->atom);
@@ -762,14 +753,141 @@ void ofono_radio_settings_register(struct ofono_radio_settings *rs)
 					NULL, rs, NULL)) {
 		ofono_error("Could not create %s interface",
 				OFONO_RADIO_SETTINGS_INTERFACE);
-
 		return;
 	}
 
 	ofono_modem_add_interface(modem, OFONO_RADIO_SETTINGS_INTERFACE);
+
 	__ofono_atom_register(rs->atom, radio_settings_unregister);
 }
 
+static void radio_mode_set_callback_at_reg(const struct ofono_error *error,
+						void *data)
+{
+	struct ofono_radio_settings *rs = data;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
+		DBG("Error setting radio access mode register time");
+
+	/*
+	 * Continue with atom register even if request fail at modem
+	 */
+	ofono_radio_finish_register(rs);
+}
+
+static void radio_band_set_callback_at_reg(const struct ofono_error *error,
+						void *data)
+{
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
+		DBG("Error setting radio access mode register time");
+	/*
+	 * Continue with atom register even if request fail at modem
+	 * ofono_radio_finish_register called by radio_mode_set_callback_at_reg
+	 */
+}
+
+static void radio_load_settings(struct ofono_radio_settings *rs,
+					const char *imsi)
+{
+	GError *error;
+
+	rs->settings = storage_open(imsi, SETTINGS_STORE);
+
+	/*
+	 * If no settings present or error; Set default.
+	 * Default RAT mode: ANY (LTE > UMTS > GSM)
+	 */
+	if (rs->settings == NULL) {
+		DBG("radiosetting storage open failed");
+		rs->mode = OFONO_RADIO_ACCESS_MODE_ANY;
+		rs->band_gsm = OFONO_RADIO_BAND_GSM_ANY;
+		rs->band_umts = OFONO_RADIO_BAND_UMTS_ANY;
+		return;
+	}
+
+	rs->imsi = l_strdup(imsi);
+
+	error = NULL;
+	rs->band_gsm = g_key_file_get_integer(rs->settings, SETTINGS_GROUP,
+					"GsmBand", &error);
+
+	if (error || radio_band_gsm_to_string(rs->band_gsm) == NULL) {
+		rs->band_gsm = OFONO_RADIO_BAND_GSM_ANY;
+		g_key_file_set_integer(rs->settings, SETTINGS_GROUP,
+						"GsmBand", rs->band_gsm);
+	}
+
+	if (error) {
+		g_error_free(error);
+		error = NULL;
+	}
+
+	rs->pending_band_gsm = rs->band_gsm;
+
+	rs->band_umts = g_key_file_get_integer(rs->settings, SETTINGS_GROUP,
+					"UmtsBand", &error);
+
+	if (error || radio_band_umts_to_string(rs->band_umts) == NULL) {
+		rs->band_umts = OFONO_RADIO_BAND_UMTS_ANY;
+		g_key_file_set_integer(rs->settings, SETTINGS_GROUP,
+						"UmtsBand", rs->band_umts);
+	}
+
+	if (error) {
+		g_error_free(error);
+		error = NULL;
+	}
+
+	rs->pending_band_umts = rs->band_umts;
+
+	rs->mode = g_key_file_get_integer(rs->settings, SETTINGS_GROUP,
+					"TechnologyPreference", &error);
+
+	if (error || radio_access_mode_to_string(rs->mode) == NULL) {
+		rs->mode = OFONO_RADIO_ACCESS_MODE_ANY;
+		g_key_file_set_integer(rs->settings, SETTINGS_GROUP,
+					"TechnologyPreference", rs->mode);
+	}
+
+	if (error) {
+		g_error_free(error);
+		error = NULL;
+	}
+
+	DBG("TechnologyPreference: %d", rs->mode);
+	DBG("GsmBand: %d", rs->band_gsm);
+	DBG("UmtsBand: %d", rs->band_umts);
+}
+
+void ofono_radio_settings_register(struct ofono_radio_settings *rs)
+{
+	struct ofono_modem *modem = __ofono_atom_get_modem(rs->atom);
+	struct ofono_sim *sim = __ofono_atom_find(OFONO_ATOM_TYPE_SIM, modem);
+
+	if (sim == NULL)
+		goto finish;
+
+	radio_load_settings(rs, ofono_sim_get_imsi(sim));
+
+	if (rs->driver->set_band != NULL)
+		rs->driver->set_band(rs, rs->band_gsm, rs->band_umts,
+					radio_band_set_callback_at_reg, rs);
+
+	if (rs->driver->set_rat_mode == NULL)
+		goto finish;
+
+	/*
+	 * Diff callback used. No need of using DBUS pending concept.
+	 * As its atom registration time - no DBUS clients.
+	 */
+	rs->driver->set_rat_mode(rs, rs->mode,
+					radio_mode_set_callback_at_reg, rs);
+	return;
+
+finish:
+	ofono_radio_finish_register(rs);
+}
+
 void ofono_radio_settings_remove(struct ofono_radio_settings *rs)
 {
 	__ofono_atom_free(rs->atom);
@@ -785,3 +903,9 @@ void *ofono_radio_settings_get_data(struct ofono_radio_settings *rs)
 {
 	return rs->driver_data;
 }
+
+struct ofono_modem *ofono_radio_settings_get_modem(
+					struct ofono_radio_settings *rs)
+{
+	return __ofono_atom_get_modem(rs->atom);
+}
